/*
 * wifimngr.c - provides "wifi" ubus object
 *
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Anjan Chanda <anjan.chanda@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <net/if.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>
#include <wifi.h>

#include "wifimngr.h"
#include "version.h"

#define WIFI_DEV_MAX_ANTENNA_NUM	4

/* if IF_OPER_* not defined */
#ifndef IF_OPER_UNKNOWN
enum {
	IF_OPER_UNKNOWN,
	IF_OPER_NOTPRESENT,
	IF_OPER_DOWN,
	IF_OPER_LOWERLAYERDOWN,
	IF_OPER_TESTING,
	IF_OPER_DORMANT,
	IF_OPER_UP,
};
#endif /* IF_OPER_UNKNOWN */

const char *standard_str[] = {
	"b",
	"g",
	"a",
	"n",
	"ac",
	"unknown"
};

#if 0	// TODO: new
const char *auth_str[] = {
	"none",
	"wep",
	"psk",
	"psk2",
	"wpa",
	"wpa2",
	"unknown"
};

const char *cipher_str[] = {
	"cipher_grp",
	"wep40",
	"tkip",
	"",
	"ccmp",
	"wep104",
	"cmac",
	"cipher_nogrp",
	"gcmp",
	"gcmp256",
	"ccmp256",
	"gmac",
	"gmac256",
	"cmac256",
	"unknown"
};
#else
/* deprecate */
const char *auth_str[] = {"OPEN", "SHARED", "WEPAUTO", "WPANONE", "WPA1", \
			"WPAPSK", "WPA2", "WPA2PSK", "Unknown"};

const char *cipher_str[] = {"NONE", "WEP", "TKIP", "AES", "CCMP256", \
			"GCMP128", "GCMP256", "Unknown"};
#endif

const char *akm_str[] = {
	"none",
	"wpa",
	"psk",
	"ft-wpa",
	"ft-psk",
	"wpa-sha256",
	"psk-sha256",
	"unknown",
};


const char *operstate_str[] = {
	"up",
	"down",
	"unknown",
	"dormant",
	"notpresent",
	"lowerdown",
	"error",
};

const int32_t bw_int[] = {
	20,
	40,
	80,
	160,
	0,
	-1,
	-1,
	8080,
	-99,
};

static const char *ifstatus_str(ifstatus_t f)
{
	const char *s;

	if (f & IFF_UP) {
		s = (f & IFF_RUNNING) ? "running" : "up";
	} else {
		s = "down";
	}

	return s;
}

static char *wifi_security_str(uint32_t a, uint32_t c, uint32_t gc, char *out)
{
	char x[32] = {0};
	const char *v;
	int i, len;
	bool sep = false;
	uint32_t cip;
	int ver;

	ver = a & (WPA_VERSION2 | WPA_VERSION1);
	if (ver == (WPA_VERSION1 | WPA_VERSION2)) {
		v = "-mixed";
	} else if (ver == WPA_VERSION2) {
		v = "2";
	} else {
		v = "";
	}

	a &= 0xff;
	len = sizeof(akm_str)/sizeof(akm_str[0]);
	for (i = 0; i < len; i++) {
		if ((a & (1 << i)) && strlen(akm_str[i])) {
			sprintf(x + strlen(x), "%s%s", sep ? "/" : "",
						akm_str[i]);
			sep = true;
		}
	}

	sprintf(out, "%s%s", x, v);
	cip = (c & 0x1) ? gc : c;
	len = sizeof(cipher_str)/sizeof(cipher_str[0]);
	for (i = 0; i < len; i++) {
		if ((cip & (1 << i)) && strlen(cipher_str[i]))
			sprintf(out + strlen(out), "+%s", cipher_str[i]);
	}

	return out;
}

static char *etostr(uint32_t e, char *out, int elen, const char **arr)
{
	char *c = "";
	int i;

	for (i = 0; i < elen; i++) {
		if (e & (1 << i)) {
			sprintf(out + strlen(out), "%s%s", c, arr[i]);
			c = "/";
		}
	}

	return out;
}

static char *etostr2(uint32_t e, char *out, int elen, const char **arr, char *sep)
{
	char *c = "";
	int i;

	for (i = 0; i < elen; i++) {
		if (e & (1 << i)) {
			sprintf(out + strlen(out), "%s%s", c, arr[i]);
			c = sep ? sep : "/";
		}
	}

	return out;
}

static char *bitmap_str(uint32_t bitmap, const char *arr[], int arr_sz,
						char *out, const char *sep)
{
	const char *c = "";
	int i;

	for (i = 0; i < arr_sz; i++) {
		if (bitmap & (1 << i)) {
			sprintf(out + strlen(out), "%s%s", c, arr[i]);
			c = sep ? sep : "/";
		}
	}

	return out;
}

static const char *ubus_objname_to_ifname(struct ubus_object *obj, int radio)
{
	const char *prefix = (radio == 1) ? "wifi.radio." : "wifi.ap.";

	if (strstr(obj->name, prefix))
		return obj->name + strlen(prefix);
	else
		return "\0";
}

#define ubus_radio_to_ifname(o)	ubus_objname_to_ifname(o, 1)
#define ubus_ap_to_ifname(o)	ubus_objname_to_ifname(o, 0)
#define ubus_sta_to_ifname(o)	ubus_objname_to_ifname(o, 0)

static int uci_get_wifi_devices(char devlist[][16])
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	int n = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (!strcmp(s->type, "wifi-device")) {
			memset(devlist[n], 0, 16);
			strncpy(devlist[n], s->e.name, 15);
			if (++n == WIFI_DEV_MAX_NUM)
				break;
		}
	}

	uci_free_context(ctx);
	return n;
}

int wifimngr_get_wifi_devices(char devs[][16])
{
	return uci_get_wifi_devices(devs);
}

static int uci_get_wifi_interfaces(struct wifimngr_iface ifs[])
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	int n = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "wifi-iface")) {
			struct uci_element *x;
			struct uci_option *op;
			char tif[16] = {0};

			uci_foreach_element(&s->options, x) {
				op = uci_to_option(x);
				if (!strncmp(x->name, "ifname", 6) &&
					op->type == UCI_TYPE_STRING) {

					strncpy(tif, op->v.string, 15);
					memset(ifs[n].iface, 0, 16);
					strncpy(ifs[n].iface, op->v.string, 15);
				} else if (!strncmp(x->name, "device", 6) &&
					op->type == UCI_TYPE_STRING) {

					memset(ifs[n].device, 0, 16);
					strncpy(ifs[n].device, op->v.string, 15);
				} else if (!strncmp(x->name, "mode", 4) &&
					op->type == UCI_TYPE_STRING) {

					if (!strncmp(op->v.string, "ap", 2))
						ifs[n].mode = WIFI_AP;
					else if (!strncmp(op->v.string, "wet", 3))
						ifs[n].mode = WIFI_STA;
				}
			}
			if (++n == WIFI_IF_MAX_NUM)
				break;
		}
	}

	uci_free_context(ctx);
	return n;
}

int wifimngr_get_wifi_interfaces(struct wifimngr_iface ifs[])
{
	return uci_get_wifi_interfaces(ifs);
}

static const char *uci_get_wifi_option(char *type, char *name, const char *option)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	const char *val = NULL;

	ctx = uci_alloc_context();
	if (!ctx)
		return NULL;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return NULL;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (!strcmp(s->type, type)) {
			val = uci_lookup_option_string(ctx, s, option);
			break;
		}
	}

	uci_free_context(ctx);
	return val ? val : NULL;
}

static void wl_dump_capabilities(struct blob_buf *bb, struct wifi_caps *caps, uint8_t *bitmap)
{
	void *f;

	f = blobmsg_open_table(bb, "capabilities");
	blobmsg_add_u8(bb, "wmm", wifi_cap_isset(bitmap, WIFI_CAP_WMM) ? true : false);
	blobmsg_add_u8(bb, "apsd", wifi_cap_isset(bitmap, WIFI_CAP_APSD) ? true : false);
	blobmsg_add_u8(bb, "shortslot", wifi_cap_isset(bitmap, WIFI_CAP_SHORT_SLOT) ? true : false);
	blobmsg_add_u8(bb, "dot11h", wifi_cap_isset(bitmap, WIFI_CAP_SPECTRUM_MGMT) ? true : false);
	//blobmsg_add_u8(bb, "dot11k", wifi_cap_isset(bitmap, WIFI_CAP_RADIO_MEAS) ? true : false);

	if (!!(caps->valid & WIFI_CAP_EXT_VALID)) {
		blobmsg_add_u8(bb, "2040coex", wifi_cap_isset(bitmap, WIFI_CAP_2040_COEX) ? true : false);
		blobmsg_add_u8(bb, "psmp", wifi_cap_isset(bitmap, WIFI_CAP_PSMP) ? true : false);
		blobmsg_add_u8(bb, "proxy_arp", wifi_cap_isset(bitmap, WIFI_CAP_PROXY_ARP) ? true : false);
		blobmsg_add_u8(bb, "dot11v_btm", wifi_cap_isset(bitmap, WIFI_CAP_11V_BSS_TRANS) ? true : false);
	}

	if (!!(caps->valid & WIFI_CAP_HT_VALID)) {
		void *n;

		n = blobmsg_open_table(bb, "dot11n");
		blobmsg_add_u8(bb, "dot11n_ldpc", wifi_cap_isset(bitmap, WIFI_CAP_HT_LDPC) ? true : false);
		blobmsg_add_u8(bb, "dot11n_40", wifi_cap_isset(bitmap, WIFI_CAP_2040) ? true : false);
		blobmsg_add_u8(bb, "dot11n_ps", wifi_cap_isset(bitmap, WIFI_CAP_HT_SMPS) ? true : false);
		blobmsg_add_u8(bb, "dot11n_sgi20", wifi_cap_isset(bitmap, WIFI_CAP_SGI20) ? true : false);
		blobmsg_add_u8(bb, "dot11n_sgi40", wifi_cap_isset(bitmap, WIFI_CAP_SGI40) ? true : false);
		blobmsg_add_u8(bb, "dot11n_tx_stbc", wifi_cap_isset(bitmap, WIFI_CAP_HT_TX_STBC) ? true : false);
		blobmsg_add_u8(bb, "dot11n_rx_stbc", wifi_cap_isset(bitmap, WIFI_CAP_HT_RX_STBC) ? true : false);
		blobmsg_close_table(bb, n);
	}/* else {
		blobmsg_add_u8(bb, "dot11n", false);
	}*/

	if (!!(caps->valid & WIFI_CAP_VHT_VALID)) {
		void *ac;

		ac = blobmsg_open_table(bb, "dot11ac");
		blobmsg_add_u8(bb, "dot11ac_160", wifi_cap_isset(bitmap, WIFI_CAP_160) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_8080", wifi_cap_isset(bitmap, WIFI_CAP_8080) ? true : false);
		if (wifi_cap_isset(bitmap, WIFI_CAP_VHT_MPDU_11454))
			blobmsg_add_u32(bb, "dot11ac_mpdu_max", 11454);
		else if (wifi_cap_isset(bitmap, WIFI_CAP_VHT_MPDU_7991))
			blobmsg_add_u32(bb, "dot11ac_mpdu_max", 7991);
		else if (wifi_cap_isset(bitmap, WIFI_CAP_VHT_MPDU_3895))
			blobmsg_add_u32(bb, "dot11ac_mpdu_max", 3895);
		blobmsg_add_u8(bb, "dot11ac_sgi80", wifi_cap_isset(bitmap, WIFI_CAP_SGI80) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_sgi160", wifi_cap_isset(bitmap, WIFI_CAP_SGI160) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_rx_ldpc", wifi_cap_isset(bitmap, WIFI_CAP_VHT_RX_LDPC) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_tx_stbc", wifi_cap_isset(bitmap, WIFI_CAP_VHT_TX_STBC) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_rx_stbc_1ss", wifi_cap_isset(bitmap, WIFI_CAP_VHT_RX_STBC_1SS) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_rx_stbc_2ss", wifi_cap_isset(bitmap, WIFI_CAP_VHT_RX_STBC_2SS) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_rx_stbc_3ss", wifi_cap_isset(bitmap, WIFI_CAP_VHT_RX_STBC_3SS) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_rx_stbc_4ss", wifi_cap_isset(bitmap, WIFI_CAP_VHT_RX_STBC_4SS) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_su_beamformer", wifi_cap_isset(bitmap, WIFI_CAP_VHT_SU_BFR) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_su_beamformee", wifi_cap_isset(bitmap, WIFI_CAP_VHT_SU_BFE) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_mu_beamformer", wifi_cap_isset(bitmap, WIFI_CAP_VHT_MU_BFR) ? true : false);
		blobmsg_add_u8(bb, "dot11ac_mu_beamformee", wifi_cap_isset(bitmap, WIFI_CAP_VHT_MU_BFE) ? true : false);
		blobmsg_close_table(bb, ac);
	}/* else {
		blobmsg_add_u8(bb, "dot11ac", false);
	}*/

	if (!!(caps->valid & WIFI_CAP_RM_VALID)) {
		void *k;

		k = blobmsg_open_table(bb, "dot11k");
		blobmsg_add_u8(bb, "dot11k_link_meas", wifi_cap_isset(bitmap, WIFI_CAP_RM_LINK) ? true : false);
		blobmsg_add_u8(bb, "dot11k_nbr_report", wifi_cap_isset(bitmap, WIFI_CAP_RM_NBR_REPORT) ? true : false);
		blobmsg_add_u8(bb, "dot11k_bcn_passive", wifi_cap_isset(bitmap, WIFI_CAP_RM_BCN_PASSIVE) ? true : false);
		blobmsg_add_u8(bb, "dot11k_bcn_active", wifi_cap_isset(bitmap, WIFI_CAP_RM_BCN_ACTIVE) ? true : false);
		blobmsg_add_u8(bb, "dot11k_bcn_table", wifi_cap_isset(bitmap, WIFI_CAP_RM_BCN_TABLE) ? true : false);
		blobmsg_add_u8(bb, "dot11k_rcpi", wifi_cap_isset(bitmap, WIFI_CAP_RM_RCPI) ? true : false);
		blobmsg_add_u8(bb, "dot11k_rsni", wifi_cap_isset(bitmap, WIFI_CAP_RM_RSNI) ? true : false);
		blobmsg_close_table(bb, k);
	} /* else
		blobmsg_add_u8(bb, "dot11k", false); */

	//blobmsg_add_u8(bb, "dot11r", wifi_cap_isset(bitmap, WIFI_CAP_FT_BSS) ? true : false);
	blobmsg_close_table(bb, f);
}

static void wifi_print_radio_stats(struct blob_buf *bb,
					struct wifi_radio_stats *s)
{
	blobmsg_add_u32(bb, "tx_bytes", s->tx_bytes);
	blobmsg_add_u32(bb, "tx_packets", s->tx_pkts);
	blobmsg_add_u32(bb, "tx_error_packets", s->tx_err_pkts);
	blobmsg_add_u32(bb, "tx_dropped_packets", s->tx_dropped_pkts);

	blobmsg_add_u32(bb, "rx_bytes", s->rx_bytes);
	blobmsg_add_u32(bb, "rx_packets", s->rx_pkts);
	blobmsg_add_u32(bb, "rx_error_packets", s->rx_err_pkts);
	blobmsg_add_u32(bb, "rx_dropped_packets", s->rx_dropped_pkts);
	blobmsg_add_u32(bb, "rx_plcp_error_packets", s->rx_plcp_err_pkts);
	blobmsg_add_u32(bb, "rx_fcs_error_packets", s->rx_fcs_err_pkts);
	blobmsg_add_u32(bb, "rx_mac_error_packets", s->rx_mac_err_pkts);
	blobmsg_add_u32(bb, "rx_unknown_packets", s->rx_unknown_pkts);
}

static void wifi_print_radio_diagnostics(struct blob_buf *bb,
					struct wifi_radio_diagnostic *d)
{
	blobmsg_add_u32(bb, "channel_busy", d->channel_busy);
	blobmsg_add_u32(bb, "tx_airtime", d->tx_airtime);
	blobmsg_add_u32(bb, "rx_airtime", d->rx_airtime);
	blobmsg_add_u32(bb, "obss_airtime", d->obss_airtime);
	blobmsg_add_u32(bb, "cca_time", d->cca_time);
	blobmsg_add_u32(bb, "false_cca_count", d->false_cca_count);
}

int wl_radio_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	char std_buf2[32] = "802.11";
	char std_buf[32] = {0};
	const char *wldev;
	ifstatus_t ifs = 0;
	uint32_t channel;
	enum wifi_bw bw;
	enum wifi_band band = BAND_2;
	unsigned long maxrate;
	int noise;
	struct wifi_radio radio;
	struct wifi_opclass opclass;
	struct wifi_opclass supp_opclass[32] = {0};
	int num_opclass;
	struct blob_buf bb;
	void *c, *d, *dd;
	int i, j;


	memset(&bb, 0, sizeof(bb));
	wldev = ubus_radio_to_ifname(obj);

	wifi_get_ifstatus(wldev, &ifs);
	wifi_get_channel(wldev, &channel, &bw);	// FIXME: bw
	if (channel && channel > 14)
		band = BAND_5;

	wifi_get_bandwidth(wldev, &bw);
	if (bw < BW20 || bw > BW160) {	// FIXME: in libwifi
		bw = BW8080 + 1;	// Unknown
	}
	wifi_get_maxrate(wldev, &maxrate);
	wifi_get_noise(wldev, &noise);

	memset(&radio, 0, sizeof(struct wifi_radio));
	wifi_radio_info(wldev, &radio);

	memset(&opclass, 0, sizeof(struct wifi_opclass));
	wifi_get_curr_opclass(wldev, &opclass);

	wifi_get_supp_opclass(wldev, &num_opclass, supp_opclass);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "radio", wldev);
	blobmsg_add_u8(&bb, "isup", (ifs & IFF_UP) ? true : false);
	blobmsg_add_string(&bb, "band",
			!!(band & BAND_5) ? "5Ghz" :
			!!(band & BAND_2) ? "2.4Ghz" : "Unknown");

	sprintf(std_buf2 + strlen(std_buf2), "%s",
		etostr(radio.oper_std, std_buf, WIFI_NUM_STD, standard_str));
	blobmsg_add_string(&bb, "standard", std_buf2);

	blobmsg_add_u32(&bb, "opclass", opclass.opclass);
	blobmsg_add_u32(&bb, "channel", channel);
	blobmsg_add_u32(&bb, "bandwidth", bw_int[bw]);
	blobmsg_add_string(&bb, "channel_ext", radio.extch == EXTCH_NONE ? "none" :
					radio.extch == EXTCH_ABOVE ? "above" :
					"below");
	blobmsg_add_u32(&bb, "tx_streams", radio.tx_streams);
	blobmsg_add_u32(&bb, "rx_streams", radio.rx_streams);
	blobmsg_add_u32(&bb, "noise", noise);
	blobmsg_add_u64(&bb, "maxrate", maxrate);
	c = blobmsg_open_array(&bb, "supp_rates");
	for (i = 0; i < 32 && radio.supp_rates[i] != 0; i++) {
		blobmsg_add_u32(&bb, "", radio.supp_rates[i]);
	}

	blobmsg_close_array(&bb, c);
	c = blobmsg_open_array(&bb, "basic_rates");
	for (i = 0; i < 32 && radio.basic_rates[i] != 0; i++) {
		blobmsg_add_u32(&bb, "", radio.basic_rates[i]);
	}
	blobmsg_close_array(&bb, c);

	c = blobmsg_open_array(&bb, "supp_channels");
#if 0
	for (i = 0; i < 32 && radio.supp_channels[i] != 0; i++) {
		blobmsg_add_u32(&bb, "", radio.supp_channels[i]);
	}
#else
	for (i = 0; i < num_opclass; i++) {
		d = blobmsg_open_table(&bb, "");
		blobmsg_add_u32(&bb, "opclass", supp_opclass[i].g_opclass);
		blobmsg_add_u32(&bb, "bandwidth", bw_int[supp_opclass[i].bw]);
		blobmsg_add_u32(&bb, "txpower", supp_opclass[i].opchannel.txpower);
		dd = blobmsg_open_array(&bb, "channels");
		for (j = 0; j < supp_opclass[i].opchannel.num; j++)
			blobmsg_add_u32(&bb, "", supp_opclass[i].opchannel.ch[j]);
		blobmsg_close_array(&bb, dd);
		blobmsg_close_table(&bb, d);
	}
#endif
	blobmsg_close_array(&bb, c);

	if (radio.acs_capable) {
		c = blobmsg_open_table(&bb, "autochannel");
		blobmsg_add_u8(&bb, "enabled", radio.acs_enabled ? true : false);
		blobmsg_add_u32(&bb, "interval", radio.acs_interval);
		blobmsg_close_table(&bb, c);
	}

	blobmsg_add_u32(&bb, "txpower", radio.txpower);
	blobmsg_add_u8(&bb, "dot11h_capable", radio.dot11h_capable ? true : false);
	blobmsg_add_u8(&bb, "dot11h_enabled", radio.dot11h_enabled ? true : false);
	blobmsg_add_string(&bb, "regdomain", radio.regdomain);

	blobmsg_add_u32(&bb, "beacon_int", radio.beacon_int);
	blobmsg_add_u32(&bb, "dtim_period", radio.dtim_period);

	blobmsg_add_string(&bb, "preamble",
			radio.pr == SHORT_PREAMBLE ? "short" :
			radio.pr == LONG_PREAMBLE ? "long" : "auto");

	blobmsg_add_u32(&bb, "short_retry_limit", radio.srl);
	blobmsg_add_u32(&bb, "long_retry_limit", radio.lrl);
	blobmsg_add_u32(&bb, "frag_threshold", radio.frag);
	blobmsg_add_u32(&bb, "rts_threshold", radio.rts);
	blobmsg_add_u8(&bb, "aggr_enabled", radio.aggr_enable ? true : false);

	c = blobmsg_open_table(&bb, "stats");
	wifi_print_radio_stats(&bb, &radio.stats);
	blobmsg_close_table(&bb, c);

	c = blobmsg_open_table(&bb, "diagnostics");
	wifi_print_radio_diagnostics(&bb, &radio.diag);
	blobmsg_close_table(&bb, c);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int wl_radio_stats(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;
	struct wifi_radio_stats s;
	struct blob_buf bb;
	int ret;

	ifname = ubus_radio_to_ifname(obj);
	ret = wifi_radio_get_stats(ifname, &s);
	if (ret != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	wifi_print_radio_stats(&bb, &s);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int wifi_iface_link(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;

	ifname = ubus_ap_to_ifname(obj);
	fprintf(stderr, "%s: ifname: %s\n", __func__, ifname);

	//TODO: placeholder now

	return 0;
}

static int wifi_iface_qos(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;

	ifname = ubus_ap_to_ifname(obj);
	fprintf(stderr, "%s: ifname: %s\n", __func__, ifname);

	//TODO: placeholder now

	return 0;
}

static int wifi_iface_diag_counters(struct ubus_context *ctx,
					struct ubus_object *obj,
					struct ubus_request_data *req,
					const char *method,
					struct blob_attr *msg)
{
	const char *ifname;

	ifname = ubus_ap_to_ifname(obj);
	fprintf(stderr, "%s: ifname: %s\n", __func__, ifname);

	//TODO: placeholder now

	return 0;
}

static int wl_ap_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;
	ifstatus_t ifs = 0;
	ifopstatus_t opstatus;
	uint8_t bssid[6] = {0};
	char bssid_str[24] = {0};
	char ssid[64] = {0};
	uint32_t auth;
	uint32_t enc;
	//unsigned long rate;
	char auth_buf[32] = {0};
	struct wifi_ap ap;
	struct wifi_bss *bss = &ap.bss;
	struct blob_buf bb;
	uint32_t channel;
	enum wifi_bw bw;
	char std_buf[32] = {0};
	char std_buf2[32] = "802.11";

	memset(&bb, 0, sizeof(bb));
	memset(&ap, 0, sizeof(struct wifi_ap));

	ifname = ubus_ap_to_ifname(obj);

	wifi_get_ifstatus(ifname, &ifs);
	wifi_get_ifoperstatus(ifname, &opstatus);
	if (opstatus > IF_OPER_UP) {
		opstatus = IF_OPER_UNKNOWN;
	}

	wifi_get_channel(ifname, &channel, &bw);
	wifi_get_bandwidth(ifname, &bw);
	if (bw < BW20 || bw > BW160) {	// FIXME: in libwifi
		bw = BW8080 + 1;	// Unknown
	}

	wifi_ap_info(ifname, &ap);

	wifi_get_bssid(ifname, bssid);
	hwaddr_ntoa(bssid, bssid_str);
	wifi_get_ssid(ifname, ssid);
	//wifi_get_security(ifname, &akm, &enc, &g_enc);	// TODO: new
	//wifi_security_str(akm, cipher, g_cipher, sec_str);
	wifi_get_security(ifname, &auth, &enc);
	//wifi_get_bitrate(ifname, &rate);
	//wifi_get_noise(ifname, &noise);

	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "ifname", ifname);
	blobmsg_add_string(&bb, "status", ifstatus_str(ifs));
	//blobmsg_add_string(&bb, "operstate", operstate_str[opstatus]);
	blobmsg_add_string(&bb, "ssid", ssid);
	blobmsg_add_string(&bb, "bssid", bssid_str);
	//blobmsg_add_string(bb, "encryption", sec_str);	// TODO: new
	if (enc == CIPHER_NONE)
		blobmsg_add_string(&bb, "encryption", "DISABLED");
	else if (!!(enc & CIPHER_WEP)) {
		blobmsg_add_string(&bb, "encryption", "WEP");
	} else {
		blobmsg_add_string(&bb, "encryption",
			etostr(auth, auth_buf, AUTH_MAX, auth_str));
	}

	blobmsg_add_u32(&bb, "bandwidth", bw_int[bw]);
	//blobmsg_add_u64(&bb, "rate", rate);
	sprintf(std_buf2 + strlen(std_buf2), "%s",
		etostr(ap.bss.oper_std, std_buf, WIFI_NUM_STD, standard_str));
	blobmsg_add_string(&bb, "standard", std_buf2);
	blobmsg_add_u32(&bb, "num_stations", ap.bss.load.sta_count);
	blobmsg_add_u32(&bb, "utilization", ap.bss.load.utilization);
	blobmsg_add_u32(&bb, "adm_capacity", ap.bss.load.available);
	blobmsg_add_u8(&bb, "hidden", !ap.ssid_advertised ? true : false);
	blobmsg_add_u32(&bb, "max_stations", ap.assoclist_max);

	wl_dump_capabilities(&bb, &bss->caps, bss->cbitmap);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int wl_ap_stats(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;
	struct wifi_ap_stats s;
	//struct bss b;		// use struct ap
	struct blob_buf bb;
	int ret;

	ifname = ubus_ap_to_ifname(obj);

	ret = wifi_ap_get_stats(ifname, &s);
	if (ret != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "tx_bytes", s.tx_bytes);
	blobmsg_add_u32(&bb, "tx_packets", s.tx_pkts);
	blobmsg_add_u32(&bb, "tx_unicast_packets", s.tx_ucast_pkts);
	blobmsg_add_u32(&bb, "tx_multicast_packets", s.tx_mcast_pkts);
	blobmsg_add_u32(&bb, "tx_broadcast_packets", s.tx_bcast_pkts);
	blobmsg_add_u32(&bb, "tx_error_packets", s.tx_err_pkts);
	blobmsg_add_u32(&bb, "tx_retrans_packets", s.tx_rtx_pkts);
	blobmsg_add_u32(&bb, "tx_retrans_fail_packets", s.tx_rtx_fail_pkts);
	blobmsg_add_u32(&bb, "tx_retry_packets", s.tx_retry_pkts);
	blobmsg_add_u32(&bb, "tx_multi_retry_packets", s.tx_mretry_pkts);
	blobmsg_add_u32(&bb, "tx_dropped_packets", s.tx_dropped_pkts);
	blobmsg_add_u32(&bb, "ack_fail_packets", s.ack_fail_pkts);
	blobmsg_add_u32(&bb, "aggregate_packets", s.aggr_pkts);

	blobmsg_add_u32(&bb, "rx_bytes", s.rx_bytes);
	blobmsg_add_u32(&bb, "rx_packets", s.rx_pkts);
	blobmsg_add_u32(&bb, "rx_unicast_packets", s.rx_ucast_pkts);
	blobmsg_add_u32(&bb, "rx_multicast_packets", s.rx_mcast_pkts);
	blobmsg_add_u32(&bb, "rx_broadcast_packets", s.rx_bcast_pkts);
	blobmsg_add_u32(&bb, "rx_error_packets", s.rx_err_pkts);
	blobmsg_add_u32(&bb, "rx_dropped_packets", s.rx_dropped_pkts);
	blobmsg_add_u32(&bb, "rx_unknown_packets", s.rx_unknown_pkts);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int wl_assoclist(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	void *a, *t;
	int i;
	const char *ifname;
	unsigned char stas[768] = {0};	/* max 128 per interface */
	int nr = 128;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	ifname = ubus_ap_to_ifname(obj);

	if (wifi_get_assoclist(ifname, stas, &nr) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	a = blobmsg_open_array(&bb, "assoclist");
	for (i = 0; i < nr; i++) {
		unsigned char *sta = &stas[i*6];
		char sta_macstr[18] = {0};

		t = blobmsg_open_table(&bb, "");
		blobmsg_add_string(&bb, "wdev", ifname);
		snprintf(sta_macstr, 18, "%02x:%02x:%02x:%02x:%02x:%02x",
			sta[0], sta[1], sta[2], sta[3], sta[4], sta[5]);
		blobmsg_add_string(&bb, "macaddr", sta_macstr);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

/* detailed stainfo policy */
enum {
	STAINFO_MACADDR,    /* sta macaddress */
	__STAINFO_MAX,
};

static const struct blobmsg_policy stainfo_policy[__STAINFO_MAX] = {
	[STAINFO_MACADDR] = { .name = "sta", .type = BLOBMSG_TYPE_STRING },
};

static int wl_dump_stations(struct blob_buf *bb, const char *ifname,
				unsigned char *macaddr)
{
	void *t, *r;
	int i, j;
	struct wifi_sta sx;
	enum wifi_bw bw;
	int noise, rssi, snr;
	unsigned char stas[768] = {0};
	int nr = 64;

	if (wifi_get_assoclist(ifname, stas, &nr) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	for (i = 0; i < nr; i++) {
		uint32_t channel = 0;
		void *s;
		unsigned char *sta = &stas[i*6];
		char sta_macstr[18] = {0};

		if (macaddr && !hwaddr_is_zero(macaddr)
				&& memcmp(macaddr, sta, 6))
			continue;

		memset(&sx, 0, sizeof(sx));
		if (wifi_get_sta_info(ifname, sta, &sx))
			continue;

		wifi_get_channel(ifname, &channel, &bw);
		wifi_get_noise(ifname, &noise);
		wifi_get_bandwidth(ifname, &bw);
		rssi = sx.rssi[0];
		snr = rssi - noise;

		t = blobmsg_open_table(bb, "");
		hwaddr_ntoa(sx.macaddr, sta_macstr);
		blobmsg_add_string(bb, "macaddr", sta_macstr);
		blobmsg_add_string(bb, "wdev", ifname);
		blobmsg_add_string(bb, "frequency",
					(channel >= 36) ? "5GHz" : "2.4GHz");
		blobmsg_add_u32(bb, "rssi", rssi);
		blobmsg_add_u32(bb, "snr", snr);
		blobmsg_add_u32(bb, "idle", sx.idle_time);
		blobmsg_add_u32(bb, "in_network", sx.conn_time);
		blobmsg_add_u32(bb, "tx_airtime", sx.tx_airtime);
		blobmsg_add_u32(bb, "rx_airtime", sx.rx_airtime);
		blobmsg_add_u32(bb, "airtime", sx.airtime);
		blobmsg_add_u32(bb, "max_rate", sx.maxrate);

		s = blobmsg_open_table(bb, "status");
		blobmsg_add_u8(bb, "wmm",
			wifi_status_isset(sx.sbitmap, WIFI_STATUS_WMM) ? true : false);
		blobmsg_add_u8(bb, "ps",
			wifi_status_isset(sx.sbitmap, WIFI_STATUS_PS) ? true : false);
		blobmsg_close_table(bb, s);

		wl_dump_capabilities(bb, &sx.caps, sx.cbitmap);

		s = blobmsg_open_table(bb, "stats");
		blobmsg_add_u32(bb, "tx_total_pkts", sx.stats.tx_pkts);
		blobmsg_add_u64(bb, "tx_total_bytes", sx.stats.tx_bytes);
		blobmsg_add_u32(bb, "tx_failures", sx.stats.tx_fail_pkts);
		blobmsg_add_u32(bb, "tx_pkts_retries", sx.stats.tx_retry_pkts);
		blobmsg_add_u32(bb, "rx_data_pkts", sx.stats.rx_pkts);
		blobmsg_add_u64(bb, "rx_data_bytes", sx.stats.rx_bytes);
		blobmsg_add_u32(bb, "rx_failures", sx.stats.rx_fail_pkts);
		blobmsg_add_u32(bb, "rate_of_last_tx_pkt", sx.tx_rate.rate);
		blobmsg_add_u32(bb, "rate_of_last_rx_pkt", sx.rx_rate.rate);
		blobmsg_close_table(bb, s);

		r = blobmsg_open_array(bb, "rssi_per_antenna");
		for (j = 0; j < WIFI_DEV_MAX_ANTENNA_NUM; j++)
			blobmsg_add_u32(bb, "", sx.rssi[j]);
		blobmsg_close_array(bb, r);
		blobmsg_close_table(bb, t);
	}

	return 0;
}

static int wl_stations(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	struct blob_attr *tb[__STAINFO_MAX];
	char sta_macstr[18] = {0};
	uint8_t sta[6] = {0};
	const char *ifname;
	struct blob_buf bb;
	void *a;

	memset(&bb, 0, sizeof(bb));
	blobmsg_parse(stainfo_policy, __STAINFO_MAX, tb, blob_data(msg), blob_len(msg));

	ifname = ubus_ap_to_ifname(obj);

	if ((tb[STAINFO_MACADDR])) {
		strncpy(sta_macstr, blobmsg_data(tb[STAINFO_MACADDR]),
					sizeof(sta_macstr)-1);
		if (hwaddr_aton(sta_macstr, sta) == NULL) {
			return UBUS_STATUS_INVALID_ARGUMENT;
		}
	}

	blob_buf_init(&bb, 0);
	a = blobmsg_open_array(&bb, "stations");
	wl_dump_stations(&bb, ifname, sta);
	blobmsg_close_array(&bb, a);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

enum {
	SCAN_SSID,
	SCAN_BSSID,
	SCAN_CHANNEL,
	/* SCAN_TYPE,    // active, passsive, background */
	__SCAN_MAX,
};

static const struct blobmsg_policy wl_scan_policy[__SCAN_MAX] = {
	[SCAN_SSID] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
	[SCAN_BSSID] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
	[SCAN_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
	/* [SCAN_TYPE] = { .name = "type", .type = BLOBMSG_TYPE_INT32 }, */
};

int wl_scan(struct ubus_context *ctx, struct ubus_object *obj,
		  struct ubus_request_data *req, const char *method,
		  struct blob_attr *msg)
{
	struct blob_attr *tb[__SCAN_MAX];
	//char device[16];
	const char *ifname;
	char ssid[33] = {0};
	int ret;

	blobmsg_parse(wl_scan_policy, __SCAN_MAX, tb, blob_data(msg),
			blob_len(msg));

	ifname = ubus_radio_to_ifname(obj);

	if ((tb[SCAN_SSID])) {
		strncpy(ssid, blobmsg_data(tb[SCAN_SSID]), 32);
	}

	if (ssid[0]) {
		struct scan_param p;

		memset(&p, 0, sizeof(struct scan_param));
		strncpy(p.ssid, ssid, strlen(ssid));
		ret = wifi_scan(ifname, &p);
	} else {
		ret = wifi_scan(ifname, NULL);
	}

	return ret == 0 ? UBUS_STATUS_OK : UBUS_STATUS_UNKNOWN_ERROR;
}

enum {
	SCANRES_BSSID,
	__SCANRES_MAX,
};

static const struct blobmsg_policy wl_scanres_policy[__SCANRES_MAX] = {
	[SCANRES_BSSID] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
};

static void wl_scanresult_print(struct blob_buf *bb, void *buf, bool detail)
{
	struct wifi_bss *b = (struct wifi_bss *)buf;
	//char enc_buf[32] = {0};
	//char cpr_buf[32] = {0};
	char std_buf[32] = {0};
	char std_buf2[32] = "802.11";
	char bssid_str[18] = {0};
	char sec_str[64] = {0};

	blobmsg_add_string(bb, "ssid", (char *)b->ssid);
	hwaddr_ntoa(b->bssid, bssid_str);
	blobmsg_add_string(bb, "bssid", bssid_str);
	blobmsg_add_u32(bb, "channel", b->channel);
	blobmsg_add_u32(bb, "bandwidth", bw_int[b->curr_bw]);
	wifi_security_str(b->sec, b->enc, b->g_enc, sec_str);
	blobmsg_add_string(bb, "encryption", sec_str);
	blobmsg_add_string(bb, "band",
			!!(b->band & BAND_5) ? "5Ghz" :
			!!(b->band & BAND_2) ? "2.4Ghz" : "Unknown");
	blobmsg_add_u32(bb, "rssi", b->rssi);
	//blobmsg_add_u32(bb, "snr", b->snr);
	sprintf(std_buf2 + strlen(std_buf2), "%s",
		etostr(b->oper_std, std_buf, WIFI_NUM_STD, standard_str));
	blobmsg_add_string(bb, "standard", std_buf2);
	blobmsg_add_u32(bb, "load_stas", b->load.sta_count);
	blobmsg_add_u32(bb, "load_utilization", b->load.utilization);
	blobmsg_add_u32(bb, "load_available", b->load.available);

	if (detail) {
		struct wifi_bss_detail *bx =
					(struct wifi_bss_detail *)buf;
		char iebuf[1024] = {0};
		int i;

		for (i = 0; i < bx->ielen; i++) {
			sprintf(iebuf + strlen(iebuf), "%02x", bx->ie[i]);
		}
		blobmsg_add_string(bb, "ielist", iebuf);
	}
}

static int wl_scanresults(struct ubus_context *ctx, struct ubus_object *obj,
		  struct ubus_request_data *req, const char *method,
		  struct blob_attr *msg)
{
	char *data;
	struct wifi_bss *bsss;
	struct blob_attr *tb[__SCANRES_MAX];
	const char *device;
	void *a;
	void *t;
	int num = 128;  /* initial estimate for max APs in scanresults */
	int i;
	char bssid_str[18] = {0};
	uint8_t bssid[6] = {0};
	bool detail = false;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(bb));
	blobmsg_parse(wl_scanres_policy, __SCANRES_MAX, tb, blob_data(msg),
							blob_len(msg));

	device = ubus_radio_to_ifname(obj);

	if (tb[SCANRES_BSSID]) {
		struct wifi_bss_detail bx;

		strncpy(bssid_str, blobmsg_data(tb[SCANRES_BSSID]),
							sizeof(bssid_str)-1);
		if (hwaddr_aton(bssid_str, bssid) == NULL) {
			fprintf(stderr, "%s(): Invalid bssid format\n",
								__func__);
			return UBUS_STATUS_INVALID_ARGUMENT;
		}
		detail = true;
		memset(&bx, 0, sizeof(bx));
		if (wifi_get_bss_scan_result(device, bssid, &bx) != 0)
			return UBUS_STATUS_UNKNOWN_ERROR;

		blob_buf_init(&bb, 0);
		//t = blobmsg_open_table(&bb, "");
		wl_scanresult_print(&bb, &bx, detail);
		//blobmsg_close_table(&bb, t);
		ubus_send_reply(ctx, req, bb.head);
		blob_buf_free(&bb);

		return UBUS_STATUS_OK;
	}

	/* dump scanresults list */
	data = calloc(num, sizeof(struct wifi_bss));

	if (!data) {
		fprintf(stderr, "OOM scan results\n");
		return UBUS_STATUS_UNKNOWN_ERROR;
	}
	bsss = (struct wifi_bss *)data;
	if (wifi_get_scan_results(device, bsss, &num) != 0) {
		free(data);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	blob_buf_init(&bb, 0);
	a = blobmsg_open_array(&bb, "accesspoints");
	for (i = 0; i < num; i++) {
		t = blobmsg_open_table(&bb, "");
		wl_scanresult_print(&bb, &bsss[i], detail);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

#ifdef WIFIMNGR_DEBUG
	for (i = 0; i < num; i++) {
		fprintf(stderr, "[%3d]   %18s    %33s     %3d\n",
			i, bsss[i].bssid, bsss[i].ssid, bsss[i].channel);
	}
#endif
	free(data);
	return UBUS_STATUS_OK;
}

enum {
	ACS_INTERVAL_SECS,
	ACS_ALGO,
	ACS_NUM_SCANS,
	__ACS_MAX,
};

static const struct blobmsg_policy wl_acs_policy[__ACS_MAX] = {
	[ACS_INTERVAL_SECS] = { .name = "interval", .type = BLOBMSG_TYPE_INT32 },
	[ACS_ALGO] = { .name = "algo", .type = BLOBMSG_TYPE_INT32 },
	[ACS_NUM_SCANS] = { .name = "scans", .type = BLOBMSG_TYPE_INT32 },
};

int wl_autochannel(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *device;
	uint32_t ch;
	struct blob_attr *tb[__ACS_MAX];
	int ret;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(bb));
	blobmsg_parse(wl_acs_policy, __ACS_MAX, tb, blob_data(msg), blob_len(msg));

	device = ubus_radio_to_ifname(obj);
	//TODO: acs params

	ret = wifi_acs(device, NULL);

	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "code", ret);
	if (ret == 0) {
		enum wifi_bw bw;

		ret = wifi_get_channel(device, &ch, &bw);
		if (ret == 0)
			blobmsg_add_u32(&bb, "new_channel", ch);
		blobmsg_add_string(&bb, "status", "success");
	} else
		blobmsg_add_string(&bb, "status", "error");

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 	UBUS_STATUS_OK;
}

/* list neighbors policy */
enum {
	NBR_LIST_SSID,    /* ssid filter */
	NBR_LIST_CLIENT,  /* from beacon report by client */
	__NBR_LIST_MAX,
};

static const struct blobmsg_policy nbr_list_policy[__NBR_LIST_MAX] = {
	[NBR_LIST_SSID] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
	[NBR_LIST_CLIENT] = { .name = "client", .type = BLOBMSG_TYPE_STRING },
};

/* sta disconnect policy */
enum {
	STA_DISCONNECT_STA,   /* sta mac ':' separated */
	__STA_DISCONNECT_MAX,
};

static const struct blobmsg_policy sta_disconnect_policy[__STA_DISCONNECT_MAX] = {
	[STA_DISCONNECT_STA] = { .name = "sta", .type = BLOBMSG_TYPE_STRING }
};

/* sta monitor policy */
enum {
	STA_MONITOR_ADDR,   /* sta mac ':' separated */
	STA_MONITOR_GET,    /* 1 to get monitored data */
	__STA_MONITOR_MAX,
};

static const struct blobmsg_policy sta_monitor_policy[__STA_MONITOR_MAX] = {
	[STA_MONITOR_ADDR] = { .name = "sta", .type = BLOBMSG_TYPE_STRING },
	[STA_MONITOR_GET] = { .name = "get", .type = BLOBMSG_TYPE_INT32 },
};

/* add neighbor policy */
enum {
	NBR_ADD_BSSID,   /* bssid ':' separated */
	NBR_ADD_CHANNEL,
	NBR_ADD_BINFO,   /* bssid info; default = 0x3 (reachable) */
	NBR_ADD_REG,     /* regulatory; default = 0 */
	NBR_ADD_PHY,     /* phytype; default = 9 = PHY_VHT */
	__NBR_ADD_MAX,
};

static const struct blobmsg_policy nbr_add_policy[__NBR_ADD_MAX] = {
	[NBR_ADD_BSSID] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
	[NBR_ADD_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
	[NBR_ADD_BINFO] = { .name = "bssid_info", .type = BLOBMSG_TYPE_STRING },
	[NBR_ADD_REG] = { .name = "reg", .type = BLOBMSG_TYPE_INT32 },
	[NBR_ADD_PHY] = { .name = "phy", .type = BLOBMSG_TYPE_INT32 },
};

/* delete neighbor policy */
enum {
	NBR_DEL_BSSID,
	__NBR_DEL_MAX,
};

static const struct blobmsg_policy nbr_del_policy[__NBR_DEL_MAX] = {
	[NBR_DEL_BSSID] = { .name = "bssid", .type = BLOBMSG_TYPE_STRING },
};

/* send neighbor request policy */
enum {
	NBR_REQ_SSID,
	NBR_REQ_CLIENT,
	__NBR_REQ_MAX,
};

static const struct blobmsg_policy nbr_req_policy[__NBR_REQ_MAX] = {
	[NBR_REQ_SSID] = { .name = "ssid", .type = BLOBMSG_TYPE_STRING },
	[NBR_REQ_CLIENT] = { .name = "client", .type = BLOBMSG_TYPE_STRING },
};

/* send neighbor bss transition policy */
enum {
	NBR_TRANS_CLIENT,
	NBR_TRANS_BSSID,
	NBR_TRANS_TIMEOUT,
	__NBR_TRANS_MAX,
};

static const struct blobmsg_policy btmreq_policy[__NBR_TRANS_MAX] = {
	[NBR_TRANS_CLIENT] = { .name = "client", .type = BLOBMSG_TYPE_STRING },
	[NBR_TRANS_BSSID] = { .name = "bssid", .type = BLOBMSG_TYPE_ARRAY },
	[NBR_TRANS_TIMEOUT] = { .name = "timeout", .type = BLOBMSG_TYPE_INT32 },
};

int sta_disconnect(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__STA_DISCONNECT_MAX];
	uint8_t sta[6] = {0};
	char sta_str[18] = {0};
	const char *ifname;

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(sta_disconnect_policy, __STA_DISCONNECT_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[STA_DISCONNECT_STA])) {
		fprintf(stderr, "%s(): sta mac not specified!\n", __func__);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	strncpy(sta_str, blobmsg_data(tb[STA_DISCONNECT_STA]), sizeof(sta_str)-1);
	if (hwaddr_aton(sta_str, sta) == NULL) {
		fprintf(stderr, "%s(): Invalid address format. Use 00:10:22:..\n", __func__);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}
	if (wifi_disconnect_sta(ifname, sta) != 0) {
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	return UBUS_STATUS_OK;
}

int sta_monitor(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__STA_MONITOR_MAX];
	struct blob_buf bb;
	void *t, *a;
	const char *ifname;
	char sta_str[18] = {0};
	uint8_t sta[6] = {0};
	struct monitor_sta_ioreq {
		uint8_t addr[6];
		char rssi[4];
		int last_seen;
	} sta_monitor_data, *stamon = NULL;


	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(sta_monitor_policy, __STA_MONITOR_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[STA_MONITOR_ADDR])) {
		fprintf(stderr, "%s(): sta mac not specified!\n", __func__);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	strncpy(sta_str, blobmsg_data(tb[STA_MONITOR_ADDR]), sizeof(sta_str)-1);
	memset(&sta_monitor_data, 0, sizeof(sta_monitor_data));
	if (hwaddr_aton(sta_str, sta) == NULL) {
		fprintf(stderr, "%s(): Invalid address\n", __func__);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if ((tb[STA_MONITOR_GET]))
		stamon = &sta_monitor_data;

	if (wifi_monitor_sta(ifname, sta, stamon) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	if (stamon) {
		int i, tmp = 0;

		memset(&bb, 0, sizeof(bb));
		blob_buf_init(&bb, 0);
		sprintf(sta_str, "%02x:%02x:%02x:%02x:%02x:%02x",
				stamon->addr[0], stamon->addr[1],
				stamon->addr[2], stamon->addr[3],
				stamon->addr[4], stamon->addr[5]);
		t = blobmsg_open_table(&bb, sta_str);
		blobmsg_add_u32(&bb, "seen", stamon->last_seen);
		a = blobmsg_open_array(&bb, "rssi");
		for (i = 0; i < 4; i++) {
			tmp = (stamon->rssi[i] > 0) ?
				stamon->rssi[i] - 256 : stamon->rssi[i];
			blobmsg_add_u32(&bb, "", tmp);
		}
		blobmsg_close_array(&bb, a);
		blobmsg_close_table(&bb, t);

		ubus_send_reply(ctx, req, bb.head);
		blob_buf_free(&bb);
	}

	return UBUS_STATUS_OK;
}

int nbr_add(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__NBR_ADD_MAX];
	const char *ifname;
	char bssid_str[18] = {0};
	uint8_t bssid[6] = {0};
	char binfo_str[12] = {0};
	unsigned int bssid_info = 0;
	unsigned int ch = 0;
	unsigned int phy = 9;
	unsigned int reg = 0;
	struct nbr nbr;

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(nbr_add_policy, __NBR_ADD_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[NBR_ADD_BSSID])) {
		fprintf(stderr, "%s(): Neighbor Bssid not specified!\n", __func__);
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	strncpy(bssid_str, blobmsg_data(tb[NBR_ADD_BSSID]), sizeof(bssid_str)-1);
	if (hwaddr_aton(bssid_str, bssid) == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[NBR_ADD_BINFO]) {
		strncpy(binfo_str, blobmsg_data(tb[NBR_ADD_BINFO]), sizeof(binfo_str)-1);
		bssid_info = atoi(binfo_str);
	}

	if (tb[NBR_ADD_CHANNEL])
		ch = blobmsg_get_u32(tb[NBR_ADD_CHANNEL]);

	if (tb[NBR_ADD_REG])
		reg = blobmsg_get_u32(tb[NBR_ADD_REG]);

	if (tb[NBR_ADD_PHY])
		phy = blobmsg_get_u32(tb[NBR_ADD_PHY]);

	/* fprintf(stderr, "Add: %02x:%02x:%02x:%02x:%02x:%02x Ch = %d Phy = %d\n",
			bssid[0], bssid[1], bssid[2],
			bssid[3], bssid[4], bssid[5],
			ch, phy); */

	memset(&nbr, 0, sizeof(struct nbr));
	memcpy(nbr.bssid, bssid, 6);
	nbr.bssid_info = bssid_info;
	nbr.channel = ch;
	nbr.phy = phy;
	nbr.reg = reg;

	if (wifi_add_neighbor(ifname, nbr) != 0) {
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	return UBUS_STATUS_OK;
}

int nbr_del(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__NBR_DEL_MAX];
	const char *ifname;
	char bssid_str[18] = {0};
	uint8_t bssid[6] = {0};

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(nbr_del_policy, __NBR_DEL_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[NBR_DEL_BSSID]))
		return UBUS_STATUS_INVALID_ARGUMENT;

	strncpy(bssid_str, blobmsg_data(tb[NBR_DEL_BSSID]), sizeof(bssid_str)-1);
	if (hwaddr_aton(bssid_str, bssid) == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (wifi_del_neighbor(ifname, bssid) != 0) {
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	return UBUS_STATUS_OK;
}

static int _nbr_list(struct blob_buf *bb, const char *ifname, bool wildcard)
{
	char bssid_str[18] = {0};
	struct nbr *nbr;
	int nr = 32;
	void *a, *t;
	int ret;
	int i;

	nbr = calloc(nr, sizeof(struct nbr));
	if (!nbr) {
		fprintf(stderr, "OOM get_neighbor_list()\n");
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	ret = wifi_get_neighbor_list(ifname, nbr, &nr);
	if (ret)
		goto out_exit;

	if (!wildcard)
		a = blobmsg_open_array(bb, "neighbors");

	for (i = 0; i < nr; i++) {
		struct nbr *e;

		e = nbr + i;
		t = blobmsg_open_table(bb, "");
		sprintf(bssid_str, "%02X:%02X:%02X:%02X:%02X:%02X",
				e->bssid[0], e->bssid[1], e->bssid[2],
				e->bssid[3], e->bssid[4], e->bssid[5]);
		blobmsg_add_string(bb, "bssid", bssid_str);
		blobmsg_add_u32(bb, "bss_info", e->bssid_info);
		blobmsg_add_u32(bb, "regulatory", e->reg);
		blobmsg_add_u32(bb, "channel", e->channel);
		blobmsg_add_u32(bb, "phy", e->phy);
		blobmsg_close_table(bb, t);
	}

	if (!wildcard)
		blobmsg_close_array(bb, a);

out_exit:
	free(nbr);
	return ret;
}

static int _sta_nbr_list(struct blob_buf *bb, const char *ifname,
						unsigned char *sta)
{
	struct sta_nbr *snbr;
	int nr = 32;
	void *a, *t;
	char bssid_str[18] = {0};
	int ret;
	int i;

	snbr = calloc(nr, sizeof(struct sta_nbr));
	if (!snbr) {
		fprintf(stderr, "OOM get_neighbor_list()\n");
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	ret = wifi_get_beacon_report(ifname, sta, snbr, &nr);
	if (ret)
		goto out_exit;

	a = blobmsg_open_array(bb, "neighbors");
	for (i = 0; i < nr; i++) {
		struct sta_nbr *e;

		e = snbr + i;
		t = blobmsg_open_table(bb, "");
		sprintf(bssid_str, "%02X:%02X:%02X:%02X:%02X:%02X",
				e->bssid[0], e->bssid[1], e->bssid[2],
				e->bssid[3], e->bssid[4], e->bssid[5]);
		blobmsg_add_string(bb, "bssid", bssid_str);
		blobmsg_add_u32(bb, "rssi", e->rssi);
		blobmsg_add_u32(bb, "rsni", e->rsni);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, a);

out_exit:
	free(snbr);
	return ret;
}

static int nbr_list(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__NBR_LIST_MAX];
	char sta_macstr[18] = {0};
	uint8_t sta[6] = {0};
	struct blob_buf bb;
	const char *ifname;

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(nbr_list_policy, __NBR_LIST_MAX, tb,
					blob_data(msg), blob_len(msg));

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	if ((tb[NBR_LIST_CLIENT])) {
		strncpy(sta_macstr, blobmsg_data(tb[NBR_LIST_CLIENT]),
						sizeof(sta_macstr)-1);
		if (hwaddr_aton(sta_macstr, sta) == NULL)
			return UBUS_STATUS_INVALID_ARGUMENT;

		if (_sta_nbr_list(&bb, ifname, sta) != 0)
			return UBUS_STATUS_UNKNOWN_ERROR;
	} else {
		if (_nbr_list(&bb, ifname, false) != 0)
			return UBUS_STATUS_UNKNOWN_ERROR;
	}

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return UBUS_STATUS_OK;
}

int nbr_request(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__NBR_REQ_MAX];
	const char *ifname;
	char da_macstr[18] = {0};
	uint8_t da[6] = {0};

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(nbr_req_policy, __NBR_REQ_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[NBR_REQ_CLIENT]))
		return UBUS_STATUS_INVALID_ARGUMENT;

	strncpy(da_macstr, blobmsg_data(tb[NBR_REQ_CLIENT]), sizeof(da_macstr)-1);
	if (hwaddr_aton(da_macstr, da) == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (wifi_req_beacon_report(ifname, da) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	return UBUS_STATUS_OK;
}

int nbr_transition(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *req, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__NBR_TRANS_MAX];
	const char *ifname;
	char sta_macstr[18] = {0};
	uint8_t sta[6] = {0};
	uint8_t t_bss[6] = {0};
	char bss_str[18] = {0};
	unsigned int tmo = 0;
	struct blob_attr *attr;
	int rem;
	int i = 0;
	unsigned char bss[72] = {0};

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(btmreq_policy, __NBR_TRANS_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[NBR_TRANS_CLIENT]))
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[NBR_TRANS_BSSID]) {
		blobmsg_for_each_attr(attr, tb[NBR_TRANS_BSSID], rem) {
			if (blobmsg_type(attr) != BLOBMSG_TYPE_STRING)
				continue;

			strncpy(bss_str, blobmsg_data(attr), sizeof(bss_str)-1);
			if (hwaddr_aton(bss_str, t_bss) == NULL)
				return UBUS_STATUS_INVALID_ARGUMENT;

			memcpy(&bss[i*6], t_bss, 6);

			// blobmsg_name(attr),
			// blobmsg_data(attr),
			// blobmsg_data_len(attr));
			if (++i > 9)
				break;
		}
	}

	if (tb[NBR_TRANS_TIMEOUT])
		tmo = blobmsg_get_u32(tb[NBR_TRANS_TIMEOUT]);

	strncpy(sta_macstr, blobmsg_data(tb[NBR_TRANS_CLIENT]), sizeof(sta_macstr)-1);
	if (hwaddr_aton(sta_macstr, sta) == NULL)
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (wifi_req_bss_transition(ifname, sta, i, bss, tmo) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	return UBUS_STATUS_OK;
}

/* vendor_ie add/del policy */
enum {
	VSIE_MGMT_STYPE,       /* bitmask of (1 << mgmt frame subtype) */
	VSIE_OUI,
	VSIE_DATA,
	__VSIE_MAX,
};

static const struct blobmsg_policy vsie_policy[__VSIE_MAX] = {
	[VSIE_MGMT_STYPE] = { .name = "mgmt", .type = BLOBMSG_TYPE_INT32 },
	[VSIE_OUI] = { .name = "oui", .type = BLOBMSG_TYPE_STRING},
	[VSIE_DATA] = { .name = "data", .type = BLOBMSG_TYPE_STRING},
};

int vsie_add(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *ureq, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__VSIE_MAX];
	const char *ifname;
	uint8_t vsie_buf[256] = {0};
	struct vendor_iereq *req = (struct vendor_iereq *)vsie_buf;
	uint8_t *data = req->ie.data;
	char data_str[499] = {0};	/* (256 - 7) * 2 + 1 */
	char oui_str[7] = {0};          /* 3 * 2 + 1 */
	uint8_t oui[3] = {0};
	int data_len = 0;
	int data_strlen;

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(vsie_policy, __VSIE_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[VSIE_OUI]) || !(tb[VSIE_DATA]))
		return UBUS_STATUS_INVALID_ARGUMENT;

	req->mgmt_subtype = 0;
	req->ie.ie_hdr.eid = 0xdd;
	req->ie.ie_hdr.len = 0;

	if (tb[VSIE_MGMT_STYPE])
		req->mgmt_subtype = blobmsg_get_u32(tb[VSIE_MGMT_STYPE]);

	if (blobmsg_data_len(tb[VSIE_OUI]) != sizeof(oui_str))
		return UBUS_STATUS_INVALID_ARGUMENT;

	strncpy(oui_str, blobmsg_data(tb[VSIE_OUI]), 6);
	strtob(oui_str, 7, oui);
	memcpy(req->ie.oui, oui, 3);
	req->ie.ie_hdr.len += 3;

	data_strlen = blobmsg_data_len(tb[VSIE_DATA]);
	if (data_strlen > sizeof(data_str) - 1)
		return UBUS_STATUS_INVALID_ARGUMENT;

	strncpy(data_str, blobmsg_data(tb[VSIE_DATA]), data_strlen);
	strtob(data_str, data_strlen, &data[2]);

	data_len = (data_strlen - 1) / 2;
	data[0] = 0xdd;
	data[1] = data_len;
	req->ie.ie_hdr.len += 2 + data_len;



	if (wifi_add_vendor_ie(ifname, req) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	return UBUS_STATUS_OK;
}

int vsie_del(struct ubus_context *ctx, struct ubus_object *obj,
		      struct ubus_request_data *ureq, const char *method,
		      struct blob_attr *msg)
{
	struct blob_attr *tb[__VSIE_MAX];
	const char *ifname;
	uint8_t vsie_buf[256] = {0};
	struct vendor_iereq *req = (struct vendor_iereq *)vsie_buf;
	uint8_t *data = req->ie.data;
	char data_str[499] = {0};	/* (256 - 7) * 2 + 1 */
	char oui_str[7] = {0};          /* 3 * 2 + 1 */
	uint8_t oui[3] = {0};
	int data_strlen;
	int data_len = 0;
	//uint32_t mgmt_frames = 0;

	ifname = ubus_ap_to_ifname(obj);
	blobmsg_parse(vsie_policy, __VSIE_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (!(tb[VSIE_OUI]))
		return UBUS_STATUS_INVALID_ARGUMENT;

	req->mgmt_subtype = 0;
	req->ie.ie_hdr.eid = 0xdd;
	req->ie.ie_hdr.len = 0;

	if (tb[VSIE_MGMT_STYPE])
		req->mgmt_subtype = blobmsg_get_u32(tb[VSIE_MGMT_STYPE]);


	if (blobmsg_data_len(tb[VSIE_OUI]) != sizeof(oui_str))
		return UBUS_STATUS_INVALID_ARGUMENT;

	strncpy(oui_str, blobmsg_data(tb[VSIE_OUI]), 6);
	strtob(oui_str, 7, oui);
	memcpy(req->ie.oui, oui, 3);
	req->ie.ie_hdr.len += 3;

	if (tb[VSIE_DATA]) {
		/* match data starting with pattern */
		data_strlen = blobmsg_data_len(tb[VSIE_DATA]);
		if (data_strlen > sizeof(data_str) - 1)
			return UBUS_STATUS_INVALID_ARGUMENT;

		strncpy(data_str, blobmsg_data(tb[VSIE_DATA]), data_strlen);
		strtob(data_str, data_strlen, &data[2]);

		data_len = (data_strlen - 1) / 2;
		data[0] = 0xdd;
		data[1] = data_len;
		req->ie.ie_hdr.len += 2 + data_len;
	}

	if (wifi_del_vendor_ie(ifname, req) != 0)
		return UBUS_STATUS_UNKNOWN_ERROR;

	return UBUS_STATUS_OK;
}

static int wl_sta_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;
	ifstatus_t ifs = 0;
	struct wifi_sta sta;
	struct blob_buf bb;
	int ret;

	ifname = ubus_sta_to_ifname(obj);
	wifi_get_ifstatus(ifname, &ifs);

	memset(&sta, 0, sizeof(struct wifi_sta));
	ret = wifi_sta_info(ifname, &sta);
	if (ret)
		return UBUS_STATUS_UNKNOWN_ERROR;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "ifname", ifname);
	blobmsg_add_string(&bb, "status", ifstatus_str(ifs));
	// TODO: other attrs
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int wl_sta_stats(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	const char *ifname;
	struct wifi_sta_stats s;
	struct blob_buf bb;
	int ret;

	ifname = ubus_sta_to_ifname(obj);

	ret = wifi_sta_get_stats(ifname, &s);
	if (ret)
		return UBUS_STATUS_UNKNOWN_ERROR;

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);
	// TODO
	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

/* disconnect from ap policy */
enum {
	AP_DISCONNECT_REASON,   /* deauth/disassoc reason code */
	__AP_DISCONNECT_MAX,
};

static const struct blobmsg_policy ap_disconnect_policy[__AP_DISCONNECT_MAX] = {
	[AP_DISCONNECT_REASON] = { .name = "reason", .type = BLOBMSG_TYPE_INT32}
};

static int wl_sta_disconnect_ap(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_attr *tb[__AP_DISCONNECT_MAX];
	unsigned int reason_code = 0;
	const char *ifname;
	int ret;

	ifname = ubus_sta_to_ifname(obj);
	blobmsg_parse(ap_disconnect_policy, __AP_DISCONNECT_MAX, tb,
					blob_data(msg), blob_len(msg));

	if (tb[AP_DISCONNECT_REASON])
		reason_code = blobmsg_get_u32(tb[AP_DISCONNECT_REASON]);

	ret = wifi_sta_disconnect_ap(ifname, reason_code);
	if (ret)
		return UBUS_STATUS_UNKNOWN_ERROR;

	return UBUS_STATUS_OK;
}

static int wl_status(struct ubus_context *ctx, struct ubus_object *obj,
		struct ubus_request_data *req, const char *method,
		struct blob_attr *msg)
{
	char radios[WIFI_DEV_MAX_NUM][16] = {0};
	struct wifimngr_iface ifs[WIFI_IF_MAX_NUM];
	struct blob_buf bb;
	int num_radios = 0;
	int num_ifs = 0;
	void *t, *tt, *ttt, *tttt, *t5;
	int i;

	num_radios = wifimngr_get_wifi_devices(radios);
	if (num_radios < 0)
		return 0;

	num_ifs = wifimngr_get_wifi_interfaces(ifs);
	if (num_ifs < 0)
		return 0;

	/* fprintf(stderr, "radios = %d    ifs = %d\n", num_radios, num_ifs); */

	memset(&bb, 0, sizeof(bb));
	blob_buf_init(&bb, 0);

	t = blobmsg_open_array(&bb, "radios");
	for (i = 0; i < num_radios; i++) {
		uint32_t channel;
		uint32_t channels[32] = {0};
		ifstatus_t ifstat = 0;
		enum wifi_band band = BAND_2;
		unsigned long rate;
		int noise;
		enum wifi_bw bw;
		//int bwi = 20;
		uint8_t s = 0;
		char s_str[32] = {0};
		struct wifi_caps radiocaps;
		char alpha2[3] = {0};
		const char *country;
		int nr = 0;
		int j;


		wifi_get_ifstatus(radios[i], &ifstat);
		wifi_get_channel(radios[i], &channel, &bw);
		if (channel && channel >= 36 && channel < 200)
			band = BAND_5;

		wifi_get_bandwidth(radios[i], &bw);
		if (bw < BW20 || bw > BW160) {	// FIXME: in libwifi
			bw = BW_UNKNOWN;
		}

		if (!wifi_get_country(radios[i], alpha2))
			country = alpha2;
		else
			country = uci_get_wifi_option("wifi-device", radios[i],
								"country");

		wifi_radio_get_caps(radios[i], &radiocaps);

		wifi_get_supp_channels(radios[i], channels, &nr,
					country == NULL ? "" : country,
					band, bw);

		wifi_get_maxrate(radios[i], &rate);
		wifi_get_noise(radios[i], &noise);
		wifi_get_oper_stds(radios[i], &s);
		bitmap_str(s, standard_str, ARRAY_SIZE(standard_str), s_str, ",");

		ttt = blobmsg_open_table(&bb, "");

		blobmsg_add_string(&bb, "name", radios[i]);
		blobmsg_add_u8(&bb, "isup", (ifstat & IFF_UP) ? true : false);
		blobmsg_add_string(&bb, "standard", s_str);
		blobmsg_add_string(&bb, "country", country);
		blobmsg_add_string(&bb, "band",
			!!(band & BAND_5) ? "5Ghz" :
			!!(band & BAND_2) ? "2.4Ghz" : "Unknown");

		blobmsg_add_u32(&bb, "channel", channel);
		blobmsg_add_u32(&bb, "frequency", wifi_channel_to_freq(channel));
		blobmsg_add_u32(&bb, "bandwidth", bw_int[bw]);
		blobmsg_add_u32(&bb, "noise", noise);
		blobmsg_add_u64(&bb, "rate", rate);
		t5 = blobmsg_open_array(&bb, "channels");
		for (j = 0; j < nr && channels[j] != 0; j++) {
			blobmsg_add_u32(&bb, "", channels[j]);
		}
		blobmsg_close_array(&bb, t5);


		tt = blobmsg_open_array(&bb, "accesspoints");
		for (j = 0; j < num_ifs; j++) {
			char *ifname = ifs[j].iface;
			uint8_t bssid[6] = {0};
			char bssidstr[18] = {0};
			char ssid[64] = {0};
			//ifopstatus_t opstatus;
			uint32_t auth;
			uint32_t enc;
			struct wifi_caps apcaps;


			if (ifs[j].device[0] &&
				!strncmp(ifs[j].device, radios[i], 16) &&
				ifs[j].mode == WIFI_AP) {
				/* fprintf(stderr, "Matched: rad = %s  if = %s\n",
						ifs[j].device, ifname); */
				//memset(ifs[j].device, 0, 16);
				;
			} else {
				continue;
			}


			wifi_get_ifstatus(ifname, &ifstat);
			//wifi_get_ifoperstatus(ifname, &opstatus);
			wifi_get_bssid(ifname, bssid);
			hwaddr_ntoa(bssid, bssidstr);
			wifi_get_ssid(ifname, ssid);
			wifi_get_security(ifname, &auth, &enc);
			wifi_get_caps(ifname, &apcaps);

			tttt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "ifname", ifname);
			blobmsg_add_string(&bb, "status", ifstatus_str(ifstat));
			//blobmsg_add_string(&bb, "operstate", operstate_str[opstatus]);
			blobmsg_add_string(&bb, "ssid", ssid);
			blobmsg_add_string(&bb, "bssid", bssidstr);

			blobmsg_close_table(&bb, tttt);
		}
		blobmsg_close_array(&bb, tt);

		tt = blobmsg_open_array(&bb, "backhauls");
		for (j = 0; j < num_ifs; j++) {
			char *ifname = ifs[j].iface;
			//ifopstatus_t opstatus;

			if (ifs[j].device[0] &&
				!strncmp(ifs[j].device, radios[i], 16) &&
				ifs[j].mode == WIFI_STA) {
				;
			} else {
				continue;
			}

			wifi_get_ifstatus(ifname, &ifstat);
			//wifi_get_ifoperstatus(ifname, &opstatus);

			tttt = blobmsg_open_table(&bb, "");
			blobmsg_add_string(&bb, "ifname", ifname);
			blobmsg_add_string(&bb, "status", ifstatus_str(ifstat));
			//blobmsg_add_string(&bb, "operstate", operstate_str[opstatus]);
			blobmsg_close_table(&bb, tttt);
		}
		blobmsg_close_array(&bb, tt);

		blobmsg_close_table(&bb, ttt);
	}
	blobmsg_close_array(&bb, t);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

enum {
	GET_PNAME,
	__GET_MAX,
};

static const struct blobmsg_policy radio_get_param_policy[__GET_MAX] = {
	[GET_PNAME] = { .name = "param", .type = BLOBMSG_TYPE_STRING },
};

int wl_radio_get_param(struct ubus_context *ctx, struct ubus_object *obj,
		  struct ubus_request_data *req, const char *method,
		  struct blob_attr *msg)
{
	struct blob_attr *tb[__GET_MAX];
	const char *device;
	struct blob_buf bb;
	int data = 0;	// FIXME: data type based on 'param' name
	int len = 0;	//        also len
	int ret;

	memset(&bb, 0, sizeof(bb));
	blobmsg_parse(radio_get_param_policy, __GET_MAX, tb,
						blob_data(msg), blob_len(msg));

	device = ubus_radio_to_ifname(obj);

	ret = wifi_radio_get_param(device, "temperature", &len, &data);
	blob_buf_init(&bb, 0);
	blobmsg_add_u32(&bb, "status", ret);
	if (ret == 0) {
		blobmsg_add_u32(&bb, "value_int", data);
	}

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);
	return 	UBUS_STATUS_OK;
}

#define MAX_WIFI_METHODS	8
int add_wifi_methods(struct ubus_object *wifi_obj)
{
	int n_methods = 0;
	struct ubus_method *wifi_methods;

	wifi_methods = calloc(MAX_WIFI_METHODS, sizeof(struct ubus_method));
	if (!wifi_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(wifi_methods, n_methods,
		UBUS_METHOD_NOARG("status", wl_status));

	wifi_obj->methods = wifi_methods;
	wifi_obj->n_methods = n_methods;

	return 0;
}

int wifimngr_add_object(struct wifimngr *w,
				struct ubus_context *ctx,
				const char *objname,
				int (*add_methods)(struct ubus_object *o))
{
	struct ubus_object *wobj;
	struct ubus_object_type *wobj_type;
	int ret;

	wobj = calloc(1, sizeof(struct ubus_object));
	if (!wobj)
		return -ENOMEM;

	wobj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!wobj_type) {
		free(wobj);
		return -ENOMEM;
	}

	wobj->name = objname;
	if (add_methods)
		add_methods(wobj);

	wobj_type->name = wobj->name;
	wobj_type->n_methods = wobj->n_methods;
	wobj_type->methods = wobj->methods;
	wobj->type = wobj_type;

	ret = ubus_add_object(ctx, wobj);
	if (!ret) {
		/* fprintf(stderr, "Added '%s' radio obj\n", wobj->obj.name); */
		w->wifi = wobj;
	}

	return ret;
}

#define MAX_RADIO_METHODS	16

static int add_radio_methods(struct ubus_object *radio_obj,
				const char *radioname)
{
	struct ubus_method *radio_methods;
	int n_methods = 0;

	radio_methods = calloc(MAX_RADIO_METHODS, sizeof(struct ubus_method));
	if (!radio_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(radio_methods, n_methods,
		UBUS_METHOD_NOARG("status", wl_radio_status));

	UBUS_METHOD_ADD(radio_methods, n_methods,
			UBUS_METHOD_NOARG("stats", wl_radio_stats));

	UBUS_METHOD_ADD(radio_methods, n_methods,
		UBUS_METHOD("get", wl_radio_get_param, radio_get_param_policy));

	//UBUS_METHOD_ADD(radio_methods, n_methods,
	//	UBUS_METHOD_NOARG("temperature", wl_temperature));

//	if (libwifi_supports(radioname, "wifi_scan")) {
		UBUS_METHOD_ADD(radio_methods, n_methods,
			UBUS_METHOD("scan", wl_scan, wl_scan_policy));

		UBUS_METHOD_ADD(radio_methods, n_methods,
			UBUS_METHOD("scanresults", wl_scanresults, wl_scanres_policy));
//	}

	UBUS_METHOD_ADD(radio_methods, n_methods,
		UBUS_METHOD("autochannel", wl_autochannel, wl_acs_policy));

	radio_obj->methods = radio_methods;
	radio_obj->n_methods = n_methods;

	return 0;
}

int wifimngr_add_radio_object(struct wifimngr *w,
				struct ubus_context *ctx,
				const char *radioname)
{
	struct wifi_ubus_object *wobj;
	int ret;
	char objname[32] = {0};

	wobj = calloc(1, sizeof(struct wifi_ubus_object));
	if (!wobj)
		return -ENOMEM;

	snprintf(objname, 28, "wifi.radio.%s", radioname);
	wobj->obj.name = strdup(objname);
	add_radio_methods(&wobj->obj, radioname);

	wobj->obj_type.name = wobj->obj.name;
	wobj->obj_type.n_methods = wobj->obj.n_methods;
	wobj->obj_type.methods = wobj->obj.methods;
	wobj->obj.type = &wobj->obj_type;

	ret = ubus_add_object(ctx, &wobj->obj);
	if (!ret) {
		/* fprintf(stderr, "Added '%s' radio obj\n", wobj->obj.name); */
		list_add_tail(&wobj->list, &w->radiolist);
	}

	return ret;
}

#define MAX_AP_METHODS	16
static int add_ap_methods(struct ubus_object *interface_obj,
				const char *ifname)
{
	int n_methods = 0;
	struct ubus_method *ap_methods;

	ap_methods = calloc(MAX_AP_METHODS, sizeof(struct ubus_method));
	if (!ap_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD_NOARG("status", wl_ap_status));

	UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD_NOARG("stats", wl_ap_stats));

	UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD_NOARG("assoclist", wl_assoclist));

	UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("stations", wl_stations, stainfo_policy));

	if (libwifi_supports(ifname, "wifi_disconnect_sta")) {
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("disconnect", sta_disconnect, sta_disconnect_policy));
	}

	if (libwifi_supports(ifname, "wifi_monitor_sta")) {
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("monitor", sta_monitor, sta_monitor_policy));
	}

	if (libwifi_supports(ifname, "wifi_add_neighbor"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("add_neighbor", nbr_add, nbr_add_policy));

	if (libwifi_supports(ifname, "wifi_del_neighbor"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("del_neighbor", nbr_del, nbr_del_policy));

	if (libwifi_supports(ifname, "wifi_get_neighbor_list"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("list_neighbor", nbr_list, nbr_list_policy));

	if (libwifi_supports(ifname, "wifi_req_beacon_report"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("request_neighbor", nbr_request, nbr_req_policy));

	if (libwifi_supports(ifname, "wifi_req_bss_transition"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("request_transition", nbr_transition, btmreq_policy));

	if (libwifi_supports(ifname, "wifi_add_vendor_ie"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("add_vendor_ie", vsie_add, vsie_policy));

	if (libwifi_supports(ifname, "wifi_del_vendor_ie"))
		UBUS_METHOD_ADD(ap_methods, n_methods,
			UBUS_METHOD("del_vendor_ie", vsie_del, vsie_policy));

	interface_obj->methods = ap_methods;
	interface_obj->n_methods = n_methods;

	return 0;
}

#define MAX_STA_METHODS	8
static int add_sta_methods(struct ubus_object *interface_obj,
				const char *ifname)
{
	int n_methods = 0;
	struct ubus_method *sta_methods;

	sta_methods = calloc(MAX_STA_METHODS, sizeof(struct ubus_method));
	if (!sta_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(sta_methods, n_methods,
			UBUS_METHOD_NOARG("status", wl_sta_status));

	UBUS_METHOD_ADD(sta_methods, n_methods,
			UBUS_METHOD_NOARG("stats", wl_sta_stats));

	if (libwifi_supports(ifname, "wifi_disconnect_ap")) {
		UBUS_METHOD_ADD(sta_methods, n_methods,
			UBUS_METHOD("disconnect", wl_sta_disconnect_ap,
						ap_disconnect_policy));
	}

	interface_obj->methods = sta_methods;
	interface_obj->n_methods = n_methods;

	return 0;
}

#ifdef WIFIMNGR_GRANULAR_OBJECTS
#define MAX_INTERFACE_MGMT_METHODS	8
static int add_interface_mgmt_methods(struct ubus_object *iface_mgmt_obj,
				const char *ifname)
{
	int n_methods = 0;
	int i;
	struct ubus_method *iface_mgmt_methods;

	iface_mgmt_methods = calloc(MAX_INTERFACE_MGMT_METHODS, sizeof(struct ubus_method));
	if (!iface_mgmt_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(iface_mgmt_methods, n_methods,
			UBUS_METHOD_NOARG("link", wifi_iface_link));

	UBUS_METHOD_ADD(iface_mgmt_methods, n_methods,
			UBUS_METHOD_NOARG("qos", wifi_iface_qos));

	iface_mgmt_obj->methods = iface_mgmt_methods;
	iface_mgmt_obj->n_methods = n_methods;

	return 0;
}

#define MAX_INTERFACE_DIAG_METHODS	8
static int add_interface_diag_methods(struct ubus_object *iface_diag_obj,
				const char *ifname)
{
	int n_methods = 0;
	int i;
	struct ubus_method *iface_diag_methods;

	iface_diag_methods = calloc(MAX_INTERFACE_DIAG_METHODS, sizeof(struct ubus_method));
	if (!iface_diag_methods)
		return -ENOMEM;

	UBUS_METHOD_ADD(iface_diag_methods, n_methods,
			UBUS_METHOD_NOARG("counters", wifi_iface_diag_counters));

	iface_diag_obj->methods = iface_diag_methods;
	iface_diag_obj->n_methods = n_methods;

	return 0;
}

int wifimngr_add_interface_subobject(
			struct wifi_ubus_object *wobj,
			struct ubus_context *ctx,
			const char *ifname,
			const char *subobject_name,
			int (*add_object_methods)(struct ubus_object *o,
							const char *ifname))
{
	struct wifi_ubus_object *sobj;
	int ret;
	char objname[64] = {0};

	sobj = calloc(1, sizeof(struct wifi_ubus_object));
	if (!sobj)
		return -ENOMEM;

	snprintf(objname, 63, "wifi.ap.%s.%s", ifname, subobject_name);
	sobj->obj.name = strdup(objname);
	if (add_object_methods)
		add_object_methods(&sobj->obj, ifname);

	sobj->obj_type.name = sobj->obj.name;
	sobj->obj_type.n_methods = sobj->obj.n_methods;
	sobj->obj_type.methods = sobj->obj.methods;
	sobj->obj.type = &sobj->obj_type;

	ret = ubus_add_object(ctx, &sobj->obj);
	if (!ret)
		list_add_tail(&sobj->list, &wobj->sobjlist);

	return ret;
}
#endif /* WIFIMNGR_GRANULAR_OBJECTS */

int wifimngr_add_interface_object(struct wifimngr *w,
				struct ubus_context *ctx,
				struct wifimngr_iface *iface)
{
	struct wifi_ubus_object *wobj;
	int ret;
	char objname[32] = {0};

	wobj = calloc(1, sizeof(struct wifi_ubus_object));
	if (!wobj)
		return -ENOMEM;

	if (iface->mode == WIFI_AP) {
		snprintf(objname, 25, "wifi.ap.%s", iface->iface);
		add_ap_methods(&wobj->obj, iface->iface);
	} else if (iface->mode == WIFI_STA) {
		snprintf(objname, 31, "wifi.backhaul.%s", iface->iface);
		add_sta_methods(&wobj->obj, iface->iface);
	} else {
		/* unhandled wifi mode */
		return -EINVAL;
	}

	wobj->obj.name = strdup(objname);
	wobj->obj_type.name = wobj->obj.name;
	wobj->obj_type.n_methods = wobj->obj.n_methods;
	wobj->obj_type.methods = wobj->obj.methods;
	wobj->obj.type = &wobj->obj_type;

	ret = ubus_add_object(ctx, &wobj->obj);
	if (!ret)
		list_add_tail(&wobj->list, &w->iflist);


#ifdef WIFIMNGR_GRANULAR_OBJECTS
	INIT_LIST_HEAD(&wobj->sobjlist);

	/* create sub objects per-interface */
	wifimngr_add_interface_subobject(wobj, ctx,
					ifname,
					"mgmt",
					add_interface_mgmt_methods);

	wifimngr_add_interface_subobject(wobj, ctx,
					ifname,
					"diagnostic",
					add_interface_diag_methods);
#endif

	return ret;
}

int wifimngr_init(struct wifimngr **w)
{
	struct wifimngr *wm;

	wm = calloc(1, sizeof(struct wifimngr));
	if (!wm)
		return -ENOMEM;

	INIT_LIST_HEAD(&wm->radiolist);
	INIT_LIST_HEAD(&wm->iflist);
	*w = wm;

	return 0;
}

int wifimngr_exit(struct wifimngr *w)
{
	struct wifi_ubus_object *p, *tmp;

	if (!w)
		return -EINVAL;

	list_for_each_entry_safe(p, tmp, &w->iflist, list) {
		list_del(&p->list);
		if (p->obj.methods)
			free((void *)p->obj.methods);
		free(p);
	}

	list_for_each_entry_safe(p, tmp, &w->radiolist, list) {
		list_del(&p->list);
		if (p->obj.methods)
			free((void *)p->obj.methods);
		free(p);
	}

	list_del_init(&w->iflist);
	list_del_init(&w->radiolist);

	if (w->wifi) {
		free(w->wifi->type);
		free(w->wifi);
	}

	free(w);

	return 0;
}

void wifimngr_version(void)
{
	fprintf(stderr, "version        : %s.%s\n", wifimngr_base_version,
					wifimngr_xtra_version);
	fprintf(stderr, "libwifi version: %s\n",
					libwifi_get_version());
}
