/*
 * wifimngr_event.c - handle wifi events
 *
 * Copyright (C) 2019 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Anjan Chanda <anjan.chanda@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/prctl.h>
#include <net/if.h>

#include <libubox/blobmsg_json.h>
#include "libubus.h"

#include <easy/easy.h>
#include <wifi.h>

#define MAX_EVENT_RESPONSE_LEN		512

extern const char *ubus_socket;

static int wifimngr_ubus_event(struct ubus_context *ctx, char *message)
{
	char event[32] = {0};
	char data[256] = {0};
	struct blob_buf b;

	sscanf(message, "%s '%[^\n]s'", event, data);
	memset(&b, 0, sizeof(b));
	blob_buf_init(&b, 0);

	if (!blobmsg_add_json_from_string(&b, data)) {
		fprintf(stderr, "Failed to parse message data: %s\n", data);
		return -1;
	}

	ubus_send_event(ctx, event, b.head);
	blob_buf_free(&b);

	return 0;
}

int wifimngr_event_cb(struct event_struct *e)
{
	struct ubus_context *ctx = (struct ubus_context *)e->priv;
	struct event_response *resp = &e->resp;
	char evtbuf[512] = {0};

	switch (resp->type) {
	case WIFI_EVENT_SCAN_START:
		{
			char fmt[256] =
				"wifi.radio '{\"radio\":\"%s\", \"action\":\"scan_%s\"}'";

			sprintf(evtbuf, fmt, e->ifname, "started");
			wifimngr_ubus_event(ctx, evtbuf);
		}
		break;
	case WIFI_EVENT_SCAN_END:
		{
			char fmt[256] =
				"wifi.radio '{\"radio\":\"%s\", \"action\":\"scan_%s\"}'";

			sprintf(evtbuf, fmt, e->ifname, "finished");
			wifimngr_ubus_event(ctx, evtbuf);
		}
		break;
	default:
		/* last resort - the default event handler */
		strncpy(evtbuf, (char *)resp->data, resp->len);
		wifimngr_ubus_event(ctx, evtbuf);
		break;
	}

	return 0;
}

int wifimngr_recv_event(const char *ifname, void *handle)
{
	int err;

	for (;;) {
		err = wifi_recv_event((char *)ifname, handle);
		if (err < 0)
			fprintf(stderr, "Error: %s\n", __func__);
	}

	return 0;
}

static void wifimngr_event_sighandler(int sig)
{
	switch (sig) {
	case SIGTERM:
	case SIGINT:
		//TODO: free stuff and exit
		exit(0);
	default:
		break;
	}
}

int wifimngr_setup_event(const char *ifname, const char *family, const char *group)
{
	int pfd[2] = { -1, -1 };
	char procname[16] = {0};
	void *handle;
	int ret = 0;
	pid_t pid;
	struct event_struct event;
	struct ubus_context *ctx;

	if (pipe(pfd) != 0) {
		fprintf(stderr, "Error opening pipe! err = %d\n", errno);
		return -1;
	}

	pid = fork();
	if (pid < 0) {
		close(pfd[0]);
		close(pfd[1]);
		fprintf(stderr, "Error forking! err = %d\n", errno);
		return -1;
	}

	if (pid > 0) {
		int res;
		int stat;

		close(pfd[1]);
		if (read(pfd[0], &res, sizeof(res)) < 0) {
			close(pfd[0]);
			waitpid(pid, &stat, WNOHANG);
		}

		return pid;
	}

	snprintf(procname, 16, "%s_%s", family, group);
	prctl(PR_SET_NAME, procname);
	prctl(PR_SET_PDEATHSIG, SIGTERM);
	close(pfd[0]);

	ctx = ubus_connect(ubus_socket);
	if (!ctx) {
		fprintf(stderr, "Failed to connect to ubus\n");
		ret = -1;
		goto safe_exit;
	}

	set_sighandler(SIGTERM, wifimngr_event_sighandler);
	memset(&event, 0, sizeof(struct event_struct));
	strncpy(event.ifname, ifname, 16);
	strncpy(event.family, family, 32);
	strncpy(event.group, group, 32);
	event.priv = ctx;
	event.cb = wifimngr_event_cb;
	/* setup response buffer */
	event.resp.data = calloc(MAX_EVENT_RESPONSE_LEN, sizeof(uint8_t));
	if (event.resp.data == NULL)
		goto free_priv;

	event.resp.len = event.resp.data ? MAX_EVENT_RESPONSE_LEN : 0;

	ret = wifi_register_event((char *)ifname, &event, &handle);

safe_exit:
	/* all done; inform parent so that it can rest peacefully */
	if (write(pfd[1], &ret, sizeof(ret)) < 0) {
		close(pfd[1]);
	}

	if (ret < 0) {
		fprintf(stderr, "Child exit due to unknown family %s\n", family);
		goto free_priv;
	}

	wifimngr_recv_event(ifname, handle);

free_priv:
	if (event.resp.len)
		free(event.resp.data);

	ubus_free(ctx);
	exit(0);
}
